unit atlasclass;

interface

type
  TPolohaMesta = record
    x, y: integer;
  end;

  PCesta = ^TCesta;

  TMesto = record
    nazev: string;
    poloha: TPolohaMesta;

    PocetCest: integer;
    IndexyCest: array of integer;

    vyhodnost: extended;  // bude uchovavat cenu, vzdalenost atd., s jakou sem program dorazil
    DorazilSemZ: integer; // index mesta, odkud sem program dorazil
  end;

  TCesta = record
    iZ, iDo: integer;  // indexy do pole mest
    vzdal: integer;
    PrumRych: integer;
    spotreba: extended;
  end;


  TAtlas = class       // trida obsluhujici atlas
  public
    mesta: array of TMesto;
    cesty: array of TCesta;
    AktualniVyhodnost: extended;  // bude uchovavat aktualni stav pri prohledavani mapy

  private
    procedure ParserMest(var data: string);
    procedure ParserCest(var data: string);
    procedure ParserSpotreb(var data: string);
    function VratIndexMesta(const ANazev: string): integer;
    procedure PridejMestuCestu(iMesto: integer; cesta: integer);
    procedure _HledejNejCestu(odkud, KdeJsem, kam: integer);

  public
    procedure NactiAtlas(const soubor: string);
    procedure UvolniZdroje;
    destructor Destroy; override;

    procedure NajdiNejCestu(AZ, ADo: integer);
  end;


implementation

uses
  SysUtils, atlas1, atlas2;

const
  buf = 1000;

function MujStrToFloat(const s: string): extended;
var
  c: integer;

begin
  Val(s, Result, c);
  if c <> 0 then raise Exception.CreateFmt('%s nen slo.', [s]);
end;




//**********parserovaci funkce**************************************************
function VratPodstrukturyDat(const JmenoStruktury: string; var data: string): string;
var
  start, konec, ZacatekPodstruktury: integer;

begin
  start := Pos('<' + JmenoStruktury + '>',  data);
  konec := Pos('</' + JmenoStruktury + '>', data);

  if (start = 0) or (konec = 0 ) then
  begin
    Result := '';
    exit;
  end;

  data[start] := '*';             // pozmenime zdojova data tak, abychom priste pokracovali od dalsi polozky
  data[konec] := '*';             // (jinak by to stale nachazelu jednu a tutez polozku)


  // podstruktura dat bude zacinat az za "<...>"
  ZacatekPodstruktury := start + Length(JmenoStruktury) + 2;
  Result := Copy(data, ZacatekPodstruktury, konec - ZacatekPodstruktury);
end;
//******************************************************************************



//**********trida TAtlas********************************************************
procedure TAtlas.UvolniZdroje;
var
  i: integer;

begin
  for i := 0 to High(mesta) do mesta[i].IndexyCest := nil;
  mesta := nil;
  cesty := nil;
end;

destructor TAtlas.Destroy;
begin
  UvolniZdroje;
end;

function TAtlas.VratIndexMesta(const ANazev: string): integer;
var
  i: integer;
begin
  //Result := -1;
  for i := 0 to High(mesta) do
  begin
    if mesta[i].nazev = ANazev then
    begin
      Result := i;
      exit;
    end;
  end;

  raise Exception.CreateFmt('Msto %s nelze najt.', [ANazev]);
end;

procedure TAtlas.PridejMestuCestu(iMesto: integer; cesta: integer);
begin
  if mesta[iMesto].PocetCest + 1 > Length(mesta[iMesto].IndexyCest) then
    SetLength(mesta[iMesto].IndexyCest, Length(mesta[iMesto].IndexyCest) + buf);

  mesta[iMesto].IndexyCest[mesta[iMesto].PocetCest] := cesta;

  Inc(mesta[iMesto].PocetCest);
end;



procedure TAtlas.ParserMest(var data: string);
var
  pocet: integer;
  polozka: string;

begin
  mesta := nil;
  pocet := 0;
  repeat
    polozka := VratPodstrukturyDat('mesto', data);
    if polozka <> '' then
    begin
      if pocet + 1 > Length(mesta) then SetLength(mesta, Length(mesta) + buf);

      // vyplni jednotliva pole mesta
      with mesta[pocet] do
      begin
        nazev := VratPodstrukturyDat('nazev', polozka);
        poloha.x := StrToInt(VratPodstrukturyDat('polohaX', polozka));
        poloha.y := StrToInt(VratPodstrukturyDat('polohaY', polozka));


        PocetCest := 0;
        IndexyCest := nil;
      end;

      Inc(pocet);
    end;

  until polozka = '';

  SetLength(mesta, pocet);
end;

procedure TAtlas.ParserCest(var data: string);
var
  pocet: integer;
  polozka: string;

begin
  cesty := nil;
  pocet := 0;
  repeat
    polozka := VratPodstrukturyDat('cesta', data);
    if polozka <> '' then
    begin
      if pocet + 1 > Length(cesty) then SetLength(cesty, Length(cesty) + buf);

      // vyplni jednotliva pole cesty
      with cesty[pocet] do
      begin
        iZ := VratIndexMesta(VratPodstrukturyDat('z', polozka));
        iDo := VratIndexMesta(VratPodstrukturyDat('do', polozka));
        vzdal := StrToInt(VratPodstrukturyDat('vzdalenost', polozka));
        PrumRych := StrToInt(VratPodstrukturyDat('prumernaRychlost', polozka));
        spotreba := -1;   // zatim neznama, bude updatela pozdeji

        // rekne temto mestum o jejich spojeni
        PridejMestuCestu(iZ, pocet);
        PridejMestuCestu(iDo, pocet);
      end;

      Inc(pocet);
    end;

  until polozka = '';

  SetLength(cesty, pocet);
end;

procedure TAtlas.ParserSpotreb(var data: string);
var
  polozka: string;
  rychlost, i: integer;
  rSpotreba: extended;

begin
  repeat
    polozka := VratPodstrukturyDat('spotreba', data);
    if polozka <> '' then
    begin
      rychlost := StrToInt(VratPodstrukturyDat('rychlost', polozka));
      rSpotreba := MujStrToFloat(VratPodstrukturyDat('litru', polozka));

      // tady je vyse slibovane updatenuti spotreby na cestach
      for i := 0 to High(cesty) do
      begin
        if cesty[i].PrumRych = rychlost then cesty[i].spotreba := rSpotreba;
      end;
    end;
  until polozka = '';
end;


procedure TAtlas.NactiAtlas(const soubor: string);
var
  data, sMesta, sCesty, sSpotreby: string;
  f: file;
  i: integer;

begin
  UvolniZdroje;

  AssignFile(f, soubor);
  FileMode := fmOpenRead;
  Reset(f, 1);

  SetLength(data, FileSize(f));     // cely soubor nactu do jednoho retezce

  BlockRead(f, data[1], Length(data));

  // odseka hlavicky a tak
  data := VratPodstrukturyDat('mapa', data);

  // rozdeli data do podstruktur mesta, cesty a spotreby
  sMesta    := VratPodstrukturyDat('mesta', data);
  sCesty    := VratPodstrukturyDat('cesty', data);
  sSpotreby := VratPodstrukturyDat('spotreby', data);

  ParserMest(sMesta);
  ParserCest(sCesty);
  ParserSpotreb(sSpotreby);

  for i := 0 to High(mesta) do
  begin
    SetLength(mesta[i].IndexyCest, mesta[i].PocetCest);
  end;


  //Form1.Memo1.Lines.Clear;
  //Form1.Memo1.Lines.Add(sSpotreby);

  CloseFile(f);

  data := '';
  sMesta := '';
  sCesty := '';
  sSpotreby := '';
end;

procedure TAtlas.NajdiNejCestu(AZ, ADo: integer);
var
  i: integer;
  
begin
  for i := 0 to High(mesta) do
  begin
    mesta[i].vyhodnost := 0;
    mesta[i].DorazilSemZ := -1;
  end;

  AktualniVyhodnost := 0;
  _HledejNejCestu(-1, AZ, ADo);
end;

procedure TAtlas._HledejNejCestu(odkud, KdeJsem, kam: integer);
var
  i: integer;
  KamJedeme: integer;

begin
  if odkud = -1 then
  begin
    mesta[KdeJsem].DorazilSemZ := KdeJsem;
    mesta[KdeJsem].vyhodnost   := AktualniVyhodnost;      // = 0
  end
  else
  begin
    if mesta[KdeJsem].DorazilSemZ = -1 then
    begin
      mesta[KdeJsem].DorazilSemZ := odkud;
      mesta[KdeJsem].vyhodnost   := AktualniVyhodnost;
    end
    else
    begin
      if mesta[KdeJsem].vyhodnost > AktualniVyhodnost then
      begin
        mesta[KdeJsem].DorazilSemZ := odkud;
        mesta[KdeJsem].vyhodnost   := AktualniVyhodnost;
      end
      else exit;
    end;
  end;

  if KdeJsem = kam then exit;

  for i := 0 to High(mesta[KdeJsem].IndexyCest) do
  begin
    // najdeme z cesty cilove misto
    KamJedeme := cesty[mesta[KdeJsem].IndexyCest[i]].iZ;
    if KamJedeme = KdeJsem then KamJedeme := cesty[mesta[KdeJsem].IndexyCest[i]].iDo;

    if Form2.RadioButton1.Checked then
      AktualniVyhodnost := AktualniVyhodnost + cesty[mesta[KdeJsem].IndexyCest[i]].vzdal
    else if Form2.RadioButton2.Checked then
      AktualniVyhodnost := AktualniVyhodnost + 1 / cesty[mesta[KdeJsem].IndexyCest[i]].PrumRych
    else if Form2.RadioButton3.Checked then
      AktualniVyhodnost := AktualniVyhodnost + cesty[mesta[KdeJsem].IndexyCest[i]].spotreba;

    _HledejNejCestu(KdeJsem, KamJedeme, kam);

    if Form2.RadioButton1.Checked then
      AktualniVyhodnost := AktualniVyhodnost - cesty[mesta[KdeJsem].IndexyCest[i]].vzdal
    else if Form2.RadioButton2.Checked then
      AktualniVyhodnost := AktualniVyhodnost - 1 / cesty[mesta[KdeJsem].IndexyCest[i]].PrumRych
    else if Form2.RadioButton3.Checked then
      AktualniVyhodnost := AktualniVyhodnost - cesty[mesta[KdeJsem].IndexyCest[i]].spotreba;
  end;
end;




//**********konec tridy TAtlas**************************************************



end.
